def splitmodel(x: str) -> tuple[str, str]:
    if ':' not in x:
        return x, ''
    
    model, specs = x.split(':')
    return model, specs

def modelname(x: str) -> str:
    return splitmodel(x)[0]

def modelspecs(x: str) -> str:
    return splitmodel(x)[1]


from ollama import Client as _OllamaClient
from fundar_llms.api.interface import PlainPromptInterface, Base64
from fundar_llms.utils import get_available_vram
from typing import Optional, Any

OLLAMA_DEFAULT_OPTIONS = dict(
    mirostat=0,           # int
    mirostat_eta=0.1,     # float
    mirostat_tau=5.0,     # float
    num_ctx=2048,         # int
    repeat_last_n=64,     # int
    repeat_penalty=1.1,   # float
    temperature=0.8,      # float
    seed=0,               # int
    stop=[],              # string
    tfs_z=1,              # float
    num_predict=128,      # int
    top_k=40,             # int
    top_p=0.9,            # float
    min_p=0.0             # float
)

class OllamaClient(PlainPromptInterface, _OllamaClient):
    def list_models(self, max_vram = None): # default max vram: (3.2 * 1.074E9)
        if max_vram is None:
            max_vram, _ = get_available_vram()
        models = self.list()
        models = models['models']
        models = [
            model['name']
            for model in models
            if model['size'] <= max_vram
        ]

        return models
    
    def generate(
            self,
            model: str,
            prompt: str,
            raw: Optional[bool] = None,
            image: Optional[Base64] = None,
            suffix: Optional[str] = None,
            format: Optional[str] = None,
            system: Optional[str] = None,
            context: Optional[Any] = None,
            stream: Optional[bool] = None,
            num_ctx: Optional[int] = None,
            num_predict: Optional[int] = None,
            temperature: Optional[float] = None,
            top_k: Optional[int] = None,
            top_p: Optional[float] = None,
            *args,
            **kwargs
    ) -> Any:
        options = dict(
            temperature = temperature,
            top_k = top_k,
            top_p = top_p,
            num_ctx = num_ctx,
            num_predict = num_predict
        )

        if not stream:
            stream = False

        options = OLLAMA_DEFAULT_OPTIONS | options # type: ignore

        return _OllamaClient.generate( # type: ignore
            self,
            model=model,
            prompt=prompt,
            system=system,
            # template=template,
            context=context,
            stream=stream,
            raw=raw,
            format=format,
            # images=images,
            options=options,
            **kwargs
        )


from typing import TypeVar, Sequence, Callable, Any, ParamSpec, TypedDict, Unpack
from collections.abc import Iterable
from functools import wraps
import inspect

P = ParamSpec('P')
T = TypeVar('T')

def flatten(nested_list: Sequence[Sequence[T]], max_level=5) -> Sequence[T]:
    return _flatten(nested_list, max_level)

def _flatten(nested_list: Sequence[T | Sequence[T]], current_level) -> Sequence[T]:
    if current_level == 0:
        return nested_list # type: ignore
    else:
        return _flatten([item # type: ignore
                         for sublist in nested_list 
                         for item in sublist # type: ignore
                         if isinstance(item, Iterable)]
                         , current_level-1)
    
def split_list_into_chunks(lst: list[T], chunk_size: int):
    return [lst[i:j] for i, j in 
            ((i, i + chunk_size) 
                for i in range(0, len(lst), chunk_size))]


def allow_opaque_constructor(**objects):
    def wrapper(f: Callable[P, T]) -> Callable[P, T]:
        sig = inspect.signature(f)

        assert all(x in sig.parameters.keys() for x in objects)

        @wraps(f)
        def _(*args, **kwargs) -> T:
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            bound_args = bound_args.arguments # type: ignore
            
            for k,v in bound_args.items(): # type: ignore
                if k not in objects:
                    continue
                if not isinstance(v, dict):
                    continue
                bound_args[k] = objects[k](**v) # type: ignore
            return f(**bound_args) # type: ignore
        return _
    return wrapper


from langchain.document_loaders import PyPDFLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_core.documents import Document
from collections.abc import Iterable, Sized
from typing import Optional
from functools import reduce
from operator import or_ as union
from typing import TypeVar, Callable
from .uuid import UuidGenerator
from .common import allow_opaque_constructor
from .cuda import is_cuda_available
import inspect


T = TypeVar('T')

DEFAULT_PDF_LOADER = PyPDFLoader

def load_document(filepath: str, loader = None, seed_id: Optional[int] = -1) -> list[Document]:
    # Un Document por página
    if not loader:
        loader = DEFAULT_PDF_LOADER
    
    assert hasattr(loader, 'load')
    
    result = loader(filepath).load()

    if seed_id != -1:
        uuid_generator = UuidGenerator(seed_id)
        for doc in result:
            doc.id = uuid_generator.next()
    
    return result
        

DEFAULT_RCT_SPLITTER = RecursiveCharacterTextSplitter(
        separators=["\n\n", "\n", "(?<=\. )", " ", "", "-\n"],
        chunk_size=1000,
        chunk_overlap=0,
        add_start_index=True
)

@allow_opaque_constructor(splitter = RecursiveCharacterTextSplitter)
def split_document(xs, splitter = None, seed_id: Optional[int] = -1) -> list[Document]:
    if not splitter:
        splitter = DEFAULT_RCT_SPLITTER

    assert hasattr(splitter, 'split_documents')
    
    if isinstance(xs, Document):
        xs = [xs]
    
    result = splitter.split_documents(xs)

    if seed_id != -1:
        uuid_generator = UuidGenerator(seed_id)

        for doc in result:
            doc.id = uuid_generator.next()
    
    return result

def load_and_split(
        filepath: str, 
        loader = None, 
        splitter = None, 
        seed_id: Optional[int] = -1,
        flatten = True,
    ):
    partial_result = load_document(filepath=filepath, 
                         loader=loader, 
                         seed_id=seed_id)
    
    result = [
        split_document(x, splitter=splitter, seed_id=seed_id)
        for x in partial_result
    ]

    if flatten:
        return [
            y
            for x in result
            for y in x
        ]
    
    return result


_sentence_transformer_obj = None
def SentenceTransformer(*args, **kwargs):
    """
    Default args:
        - model: sentence-transformers/all-mpnet-base-v2
        - device: auto (cuda if available)
    """
    global _sentence_transformer_obj
    if _sentence_transformer_obj is None:
        from sentence_transformers import SentenceTransformer # type: ignore
        _sentence_transformer_obj = SentenceTransformer

    sig = inspect.signature(_sentence_transformer_obj)

    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    bound_args = bound_args.arguments
    
    if bound_args['model_name_or_path'] is None:
        bound_args["model_name_or_path"] = 'sentence-transformers/all-mpnet-base-v2'

    device = bound_args['device'] or 'auto'

    match device:
        case 'auto':
            bound_args['device'] = 'cuda' if is_cuda_available() else 'cpu'
        case 'cuda':
            if not is_cuda_available():
                raise ValueError("CUDA is not available. Use 'cpu' instead.")
        case 'cpu':
            pass
        case _:
            raise ValueError(f"Invalid device '{device}', expected 'cpu', 'cuda' or 'auto.")

    return _sentence_transformer_obj(**bound_args)

def encode_with_multiprocessing(transformer, pool):
    assert isinstance(transformer, type(SentenceTransformer()))
    assert set(pool.keys()) == {'input', 'output', 'processes'}, f"Expected: {{'input', 'output' 'processes'}}, got = {set(pool.keys())}"
    def _(x):
        result = transformer.encode_multi_process(x, pool=pool)
        transformer.stop_multi_process_pool(pool)
        return result
    return _

@allow_opaque_constructor(sentence_transformer = SentenceTransformer)
def vectorize_document(
        x: str | Document | Iterable[Document | str], 
        sentence_transformer = None,
        uid = None, 
        additional_metadata = dict(),
        devices = None
    ):
    
    devices = devices or ['cuda' if is_cuda_available() else 'cpu']
    
    generator: UuidGenerator
    if uid is None:
        generator = UuidGenerator()

    elif isinstance(uid, UuidGenerator):
        generator = uid
    
    else:
        raise ValueError
    
    uid = generator.next()

    if sentence_transformer is None:
        sentence_transformer = SentenceTransformer()

    encode: Callable
    if len(devices) <= 1:
        encode = sentence_transformer.encode
    else:
        pool = sentence_transformer.start_multi_process_pool(devices)
        encode = encode_with_multiprocessing(sentence_transformer, pool)

    if isinstance(x, str):
        return [dict(
            page_content = x,
            metadata = {}|additional_metadata,
            id = uid,
            embedding = encode(x)
        )]
    
    if isinstance(x, Document):
        return [dict(
            page_content = x.page_content,
            metadata = x.metadata|additional_metadata,
            id = x.id,
            embedding = encode(x.page_content)
        )]

    if isinstance(x, Iterable):
        t = reduce(union, map(type, x))
        assert t in (str, Document)
        
        if t == str:
            all_documents = x
            all_metadatas: list[dict] = [dict() for _ in x]
        elif t == Document:
            all_documents, all_metadatas = \
                zip(*((y.page_content, y.metadata) 
                      for y in x 
                      if isinstance(y, Document) # my-py no detecta el tipo
                    ))
        all_ids = [x for x in generator.map(all_documents)]
        all_embeddings = encode(all_documents)

        result = [
            dict(page_content = doc_i, 
                 metadata = metadata_i, 
                 id = uid_i, 
                 embedding = embedding_i)
            for (doc_i, metadata_i, uid_i, embedding_i)
            in zip(
                all_documents,
                all_metadatas,
                all_ids,
                all_embeddings
            )
        ]

        assert ((isinstance(x, Sized) and len(result) == len(x)) or 
                (isinstance(x, Iterable) and len(result) == len([y for y in x])))
        assert len(all_ids) == len(set(all_ids)), "IDs were not unique. Please try again with a new seed."

        return result



from tqdm.auto import tqdm
from fundar_llms.dimred.umap import UMAP
from fundar_llms.dimred.tsne import TSNE
from numpy import empty as empty_vector

def umap_project_embeddings(umap_transform, embeddings, show_progress, strict: bool, dimensions: int):
    assert isinstance(umap_transform, type(UMAP()))

    if strict:
        umap_embeddings = empty_vector((len(embeddings), dimensions))
        for i,embedding in enumerate(tqdm(embeddings, disable=(not show_progress))):
            umap_embeddings[i] = umap_transform.transform([embedding])
        return umap_embeddings
    else:
        return umap_transform.transform(embeddings)

def tsne_project_embeddings(tsne_transform, embeddings, show_progress):
    assert isinstance(tsne_transform, TSNE)
    
    tsne_transform.verbose = show_progress
    result = tsne_transform.fit_transform(embeddings)
    tsne_transform.verbose = False

    return result

def project_data(transform, 
                 embeddings, by: str = None, 
                 show_progress: bool = True,
                 strict: bool = True,
                 dimensions: int = 2
                 ):
    if by is None:
        by = 'tsne' if isinstance(transform, TSNE) else 'umap'
    
    match by:
        case 'tsne':
            return tsne_project_embeddings(transform, embeddings, show_progress)
        case 'umap':
            return umap_project_embeddings(transform, embeddings, show_progress, strict, dimensions)
        case _:
            raise TypeError(f"Unknown transform {type(transform)}")


import random
from typing import Any, Optional, Iterable, Callable, ParamSpec, TypeVar
import uuid
import inspect
from types import GeneratorType
from functools import wraps
import collections
from itertools import islice

P = ParamSpec('P')
T = TypeVar('T')

def consume(iterator: Iterable[Any], n: Optional[int] = None) -> None:
    "Advance the iterator n-steps ahead. If n is none, consume entirely."
    if n is None:
        collections.deque(iterator, maxlen=0)
    else:
        next(islice(iterator, n, n), None)

def pass_generator_as_copy(*xs: Iterable[str]):
    def wrapper(f: Callable[P, T]) -> Callable[P, T]:
        sig = inspect.signature(f)

        assert all(x in sig.parameters.keys() for x in xs)

        @wraps(f)
        def _(*args, **kwargs) -> T:
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            bound_args = bound_args.arguments # type: ignore
            
            for k,v in bound_args.items(): # type: ignore
                if not k in xs:
                    continue
                if not isinstance(v, GeneratorType):
                    continue
                bound_args[k] = list(v) # type: ignore
            return f(**bound_args) # type: ignore
        return _
    return wrapper

class UuidGenerator:
    def __init__(self, seed: Optional[int] = None):
        self.seed = seed
        self._instantiate_random()

    def _instantiate_random(self):
        self.random = random.Random()

        if self.seed is not None:
            self.random.seed(self.seed)

        return self
    
    def reset(self):
        return self._instantiate_random()

    def _next(self):
        return uuid.UUID(int=self.random.getrandbits(128), version=4)

    def next(self, n: Optional[int] = None):
        if n is None:
            return self._next()
        else:
            return (self._next() for _ in range(n))
    
    def __iter__(self):
        while True:
            yield self.next()

    @pass_generator_as_copy('xs')
    def map(self, xs: Iterable[T], offset: Optional[int] = None) -> Iterable[uuid.UUID]: 
        if offset is not None:
            self.reset()

            if offset > 0:
                consume(self.next(offset))
        
        return map(lambda _: self.next(), xs)

    @pass_generator_as_copy('xs')
    def zipWith(self, xs: Iterable[T], offset: Optional[int] = None) -> Iterable[tuple[uuid.UUID, T]]:
        return zip(self.map(xs, offset=offset), xs)


from typing import Protocol, Optional, Any
from dataclasses import dataclass
from fundar_llms import Base64, Context
from fundar_llms.utils import DataclassDictUtilsMixin

response_dataclass_options = dict(
    init            = True,
    repr            = True,
    eq              = True,
    order           = False,
    unsafe_hash     = True,
    frozen          = True,
    match_args      = False,
    kw_only         = True,
    slots           = True,
    weakref_slot    = False,
)

@dataclass(**response_dataclass_options)
class BaseResponse(DataclassDictUtilsMixin):
    model: str
    prompt: str
    system: str
    response: str
    total_duration: int
    load_duration: Optional[int] = None
    done: Optional[bool] = None
    done_reason: Optional[str] = None
    context: Optional[Context] = None
    num_ctx: Optional[int] = None
    num_predict: Optional[int] = None
    temperature: Optional[float] = None
    top_k: Optional[float] = None
    top_p: Optional[float] = None
    extra: Optional[Any] = None


class PlainPromptInterface(Protocol):
    def generate(
            self,
            model: str,
            prompt: str,
            raw: Optional[bool] = None,
            image: Optional[Base64] = None,
            suffix: Optional[str] = None,
            format: Optional[str] = None,
            system: Optional[str] = None,
            context: Optional[Context] = None,
            stream: Optional[bool] = None,
            num_ctx: Optional[int] = None,
            num_predict: Optional[int] = None,
            temperature: Optional[float] = None,
            top_k: Optional[int] = None,
            top_p: Optional[float] = None,
            *args,
            **kwargs
    ) -> Any: ...





from numpy import argsort
from typing import Optional
from collections.abc import Iterable

def subdict(d, keys):
    return {k: d[k] for k in keys}

_reranker = None

def upsert_reranker_model(
        model_name='cross-encoder/ms-marco-MiniLM-L-6-v2', 
        device='cuda', 
        **kwargs
    ):

    global _reranker
    if _reranker is None:
        from sentence_transformers import CrossEncoder
        _reranker = CrossEncoder(model_name=model_name, **kwargs)
    else:
        if (_reranker.model.name_or_path != model_name) or (_reranker.model.device.type != device):
            del _reranker
            _reranker = CrossEncoder(model_name=model_name, **kwargs)
    return _reranker


_ALL_RERANK_ELEMENTS = ('elements', 'query', 'scores')
def _process_rerank(cross_encoder, universe: Iterable[str], query: str, include: Iterable[str]):
    assert all(x in _ALL_RERANK_ELEMENTS for x in include)

    pairs = [[query, x] for x in universe]
    scores_ = cross_encoder.predict(pairs)
    indices = argsort(scores_)[::-1]

    for i in indices:
        result = dict(
            elements = universe[i],
            query = query,
            scores = scores_[i]
        )

        yield tuple(subdict(result, include).values())

_INCLUDE_MAP = dict(
    all = _ALL_RERANK_ELEMENTS,
    score = ('scores', ),
    queries = ('query', ),
    element = ('elements', ),
)

def rerank(
        universe: Iterable[str], 
        query: str, 
        cross_encoder = None, 
        include: Optional[Iterable[str]] = None,
        *args,
        **kwargs,
):
    if not cross_encoder:
        cross_encoder = upsert_reranker_model()

    if include is None:
        include = ('all', )
    
    _include: tuple = tuple(_INCLUDE_MAP[x] for x in include)
    include = tuple(set(y for x in _include for y in x))

    return list(_process_rerank(
        cross_encoder=cross_encoder,
        query=query,
        universe=universe,
        include=include
    ))


from abc import ABC as AbstractBaseClass
from dataclasses import asdict
from typing import Any

class DataclassDictUtilsMixin(AbstractBaseClass):
    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        return cls(**data)
    
    def to_dict(self, compact=False) -> dict[str, Any]:
        data = asdict(self) # type: ignore
        
        if compact :
            data = {k:v for k,v in data.items() if v}
        
        return data


from fundar_llms.utils import modelname
from dataclasses import dataclass
from typing import Optional

tokenizer_model_dataclass_options = dict(
    init            = True,
    repr            = True,
    eq              = True,
    order           = False,
    unsafe_hash     = True,
    frozen          = True,
    match_args      = True,
    kw_only         = False,
    slots           = True,
    weakref_slot    = False,
)

auto_tokenizer_ = None

@dataclass(**tokenizer_model_dataclass_options)
class TokenizerModel:
    name: str
    needs_auth: bool = True

    def __str__(self):
        return self.name

    def auto_tokenizer_from_pretrained(self):
        global auto_tokenizer_
        if not auto_tokenizer_:
            from transformers import AutoTokenizer # type: ignore
            auto_tokenizer_ = AutoTokenizer
        
        if self.needs_auth:
            from dotenv import find_dotenv, dotenv_values
            from huggingface_hub import login # type: ignore
            login(dotenv_values(find_dotenv())['HUGGINGFACE_API_KEY'])
        
        return auto_tokenizer_.from_pretrained(self.name)

            
# Acá silencio al typechecker porque no detecta el __init__ generado por la dataclass.
DEFAULT_TOKENIZER_MAP = {
    'phi3.5': \
        TokenizerModel('microsoft/Phi-3.5-mini-instruct', needs_auth=False), # type: ignore
    'phi3': \
        TokenizerModel('microsoft/Phi-3-mini-128k-instruct', needs_auth=False), # type: ignore
    'llama3.1': \
        TokenizerModel('meta-llama/Llama-3.1-8B-Instruct', needs_auth=True), # type: ignore
    'llama3.2': \
        TokenizerModel('meta-llama/Llama-3.2-3B-Instruct', needs_auth=True), # type: ignore
}

def get_tokenizer(model_name: str, 
                  tokenizer_map: Optional[dict[str, TokenizerModel]] = None, 
                  default: Optional[TokenizerModel] = None) -> TokenizerModel:
    if not tokenizer_map:
        global DEFAULT_TOKENIZER_MAP
        tokenizer_map = DEFAULT_TOKENIZER_MAP
    
    if not default:
        default = tokenizer_map['phi3.5']
    
    return tokenizer_map.get(modelname(model_name), default)




from setuptools import setup, find_packages # type: ignore

setup(
    version = '1.0.4',
    name='fundar_llms',
    author='Fundar',
    description="LLM utilities created for Fundar's dev projects.",
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.10',
    setup_requires=['setuptools-git-versioning'],
    version_config={
       "dirty_template": "{tag}",
   }
)


from .tsne import *
from .umap import *
from .projections import project_data

"""
Usage:

```
from fundar_llms.dimred import TSNE, UMAP, project_data

umap_transform = UMAP(
    random_state = 0,
    transform_seed = 0,
    metric = 'cosine',
)

tsne_transform = TSNE(
    random_state = 0,
    metric = 'cosine',
)

umap_transform = umap_transform.fit(data)
tsne_transform = tsne.fit(data)

projected_umap = project_data(umap_transform, data)
projected_tsne = project_data(tsne_transform, data)
```
"""


import torch
import subprocess

_is_cuda_available_ = None
def is_cuda_available() -> bool:
    global _is_cuda_available_
    if _is_cuda_available_ is None:
        from torch.cuda import is_available as is_cuda_available_
        _is_cuda_available_ = is_cuda_available_
    return _is_cuda_available_()

def available_vram_nvidia_smi():
    try:
        result = subprocess.run(
            ['nvidia-smi', '--query-gpu=memory.total,memory.free', '--format=csv,noheader,nounits'],
            stdout=subprocess.PIPE,
            encoding='utf-8'
        )
        output = result.stdout.strip()
        total_mem, free_mem = map(int, output.split(','))
        return free_mem * 1024 * 1024, total_mem * 1024 * 1024  # Convert MB to bytes
    except FileNotFoundError:
        pass

    raise RuntimeError("nvidia-smi not found. Ensure NVIDIA drivers are installed.")

def available_vram_torch():
    if torch.cuda.is_available():
        torch.cuda.synchronize()
        total_mem = torch.cuda.get_device_properties(0).total_memory
        allocated_mem = torch.cuda.memory_allocated(0)  # Memory currently allocated by tensors
        reserved_mem = torch.cuda.memory_reserved(0)  # Memory reserved by the CUDA allocator
        free_mem = reserved_mem - allocated_mem  # Available memory from reserved pool
        return free_mem, total_mem
    else:
        raise RuntimeError("CUDA is not available.")
    
def get_available_vram():
    try:
        return available_vram_nvidia_smi()
    except Exception as ex:
        pass
    
    try:
        return available_vram_torch()
    except Exception as ex:
        pass
    
    raise RuntimeError("Coudlnt get available memory.")
